#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.

"""Module to generate and analyze persistence diagrams."""

import numpy as np
import pandas as pd

import matplotlib.pyplot as plt
from matplotlib.collections import LineCollection

from scipy.spatial.distance import pdist, cdist, squareform
from scipy.stats import gaussian_kde
from typing import Union, Optional
from typing_extensions import Literal

from .. import utils, config, core, graph


# Setup logging
logger = config.get_logger(__name__)


@utils.map_neuronlist(desc='Calc. persistence', allow_parallel=True)
def persistence_points(x: 'core.NeuronObject',
                       descriptor: Union[
                                         Literal['root_dist']
                                         ] = 'root_dist',
                       remove_cbf: bool = False
                       ) -> pd.DataFrame:
    """Calculate points for a persistence diagram.

    Based on Li et al., PLoS One (2017). Briefly, this cuts the neuron into
    linear segments, the start (birth) and end (death) of which are assigned a
    value (see `descriptor` parameter). In combination, these points represent
    a fingerprint for the topology of the neuron.

    Parameters
    ----------
    x :         TreeNeuron | MeshNeuron | NeuronList
                Neuron(s) to calculate persistence poinst for. For MeshNeurons,
                we will use the skeleton produced by/associated with its
                `.skeleton` property.
    descriptor : 'root_dist'
                Descriptor function used to calculate birth and death "time" of
                the segments:
                  - `root_dist` distance from root
    remove_cbf : bool
                In unipolar neurons (e.g. in insects) the soma is separate and
                connects to the neuron's backbone via "cell body fiber" (CBF).
                The length of the CBF can vary quite a bit. Discounting the
                CBF can make the persistence points more stable.
                If `remove_cbf=True` and the neuron has a soma (!) we ignore
                the CBF for the birth & death times. Neurons will also be
                automatically be rooted onto their soma!

    Returns
    -------
    pandas.DataFrame

    Examples
    --------
    >>> import navis
    >>> n = navis.example_neurons(1)
    >>> n.reroot(n.soma, inplace=True)
    >>> p = navis.persistence_points(n)

    References
    ----------
    Li Y, Wang D, Ascoli GA, Mitra P, Wang Y (2017) Metrics for comparing
    neuronal tree shapes based on persistent homology.
    PLOS ONE 12(8): e0182184. https://doi.org/10.1371/journal.pone.0182184

    """
    if descriptor not in ('root_dist', ):
        raise ValueError(f'Unknown "descriptor" parameter: {descriptor}')

    if isinstance(x, core.MeshNeuron):
        x = x.skeleton
    elif not isinstance(x, core.TreeNeuron):
        raise ValueError(f'Expected TreeNeuron(s), got "{type(x)}"')

    if remove_cbf and x.has_soma:
        # Reroot to soma
        x.reroot(x.soma, inplace=True)
        # Find the main branch point
        mbp = graph.find_main_branchpoint(x)

    # Generate segments
    segs = graph._generate_segments(x, weight='weight')

    # Grab starts and ends of each segment
    ends = np.array([s[0] for s in segs])
    starts = np.array([s[-1] for s in segs])

    if descriptor == 'root_dist':
        # Get geodesic distances to roots
        dist = graph.dist_to_root(x, weight='weight')
        death = np.array([dist[e] for e in ends])
        birth = np.array([dist[s] for s in starts])

        if remove_cbf and x.has_soma:
            # Subtract length of CBF
            cbf_length = graph.dist_between(x, mbp, x.soma)
            birth -= cbf_length
            death -= cbf_length

            # Drop segments that are entirely on the CBF
            starts = starts[death >= 0]
            ends = ends[death >= 0]
            birth = birth[death >= 0]
            death = death[death >= 0]

            # Clip negative births
            birth[birth < 0] = 0

    # Compile into a DataFrame
    pers = pd.DataFrame()
    pers['start_node'] = starts
    pers['end_node'] = ends
    pers['birth'] = birth
    pers['death'] = death

    return pers


def persistence_distances(q: 'core.NeuronObject',
                          t: Optional['core.NeuronObject'] = None,
                          augment: bool = True,
                          normalize: bool = True,
                          bw: float = .2,
                          **persistence_kwargs):
    """Calculate morphological similarity using persistence diagrams.

    This works by:
      1. Generate persistence points for each neuron.
      2. Create a weighted Gaussian from persistence points and sample 100
         evenly spaced points to create a feature vector.
      3. Calculate Euclidean distance.

    Parameters
    ----------
    q/t :       NeuronList
                Queries and targets, respectively. If `t=None` will run
                queries against queries. Neurons should have the same units,
                ideally nanometers.
    normalize : bool
                If True, will normalized the vector for each neuron to be within
                0-1. Set to False if the total number of linear segments matter.
    bw :        float
                Bandwidth for Gaussian kernel: larger = smoother, smaller =
                more detailed.
    augment :   bool
                Whether to augment the persistence vectors with other neuron
                properties (number of branch points & leafs and cable length).
    **persistence_kwargs
                Keyword arguments are passed to [`navis.persistence_points`][].

    Returns
    -------
    distances : pandas.DataFrame

    See Also
    --------
    [`navis.persistence_points`][]
                The function to calculate the persistence points.
    [`navis.persistence_vectors`][]
                Use this to get and inspect the actual vectors used here.

    """
    q = core.NeuronList(q)
    all_n = q

    if t:
        t = core.NeuronList(t)
        all_n += t

    # Some sanity checks
    if len(all_n) <= 1:
        raise ValueError('Need more than one neuron.')

    soma_warn = False
    root_warn = False
    for n in all_n:
        if not soma_warn:
            if n.has_soma and n.soma not in n.root:
                soma_warn = True
        if not root_warn:
            if len(n.root) > 1:
                root_warn = True

        if root_warn and soma_warn:
            break

    if soma_warn:
        logger.warning('At least some neurons are not rooted to their soma.')
    if root_warn:
        logger.warning('At least some neurons are fragmented.')

    # Get persistence points for each skeleton
    pers = persistence_points(all_n, **persistence_kwargs)

    # Get the vectors
    vectors, samples = persistence_vectors(pers, samples=100, bw=bw)

    # Normalizing the vectors will produce more useful distances
    if normalize:
        vectors = vectors / vectors.max(axis=1).reshape(-1, 1)
    else:
        vectors = vectors / vectors.max()

    if augment:
        # Collect extra data. Note that this adds only 3 more to the existing
        # 100 observations
        vec_aug = np.vstack((all_n.cable_length,
                             all_n.n_leafs,
                             all_n.n_branches)).T

        # Normalize per metric
        vec_aug = vec_aug / vec_aug.max(axis=0)

        # If we wanted to weigh those observation equal to the 100 topology
        # observations:
        # vec_aug *= 100 / vec_aug.shape[1]

        vectors = np.append(vectors, vec_aug, axis=1)

    if t:
        # Extract source and target vectors
        q_vec = vectors[:len(q)]
        t_vec = vectors[len(q):]
        return pd.DataFrame(cdist(q_vec, t_vec), index=q.id, columns=t.id)
    else:
        return pd.DataFrame(squareform(pdist(vectors)), index=q.id, columns=q.id)


def persistence_vectors(x,
                        threshold: Optional[float] = None,
                        samples: int = 100,
                        bw: float = .2,
                        center: bool = False,
                        **kwargs):
    """Produce vectors from persistence points.

    Works by creating a Gaussian and sampling `samples` evenly spaced
    points across it.

    Parameters
    ----------
    x :         navis.NeuronList | pd.DataFrame | list thereof
                The persistence points (see [`navis.persistence_points`][]).
                For vectors for multiple neurons, provide either a list of
                persistence points DataFrames or a single DataFrame with a
                "neuron_id" column.
    threshold : float, optional
                If provided, segments shorter (death - birth) than this will not
                be used to create the Gaussian.
    samples :   int
                Number of points sampled across the Gaussian.
    bw :        float
                Bandwidth for Gaussian kernel: larger = smoother, smaller =
                more detailed.
    center :    bool
                Whether to center the individual curves on their highest value.
                This is done by "rolling" the axis (using `np.roll`) which
                means that elements that roll beyond the last position are
                re-introduced at the first.

    Returns
    -------
    vectors :   np.ndarray
    samples :   np.ndarray
                Sampled distances. If `center=True` the absolute values don't
                make much sense anymore.

    References
    ----------
    Li Y, Wang D, Ascoli GA, Mitra P, Wang Y (2017) Metrics for comparing
    neuronal tree shapes based on persistent homology.
    PLOS ONE 12(8): e0182184. https://doi.org/10.1371/journal.pone.0182184

    See Also
    --------
    [`navis.persistence_points`][]
                The function to calculate the persistence points.
    [`navis.persistence_distances`][]
                Get distances based on (augmented) persistence vectors.

    """
    if isinstance(x, core.BaseNeuron):
        x = core.NeuronList(x)

    if isinstance(x, pd.DataFrame):
        pers = [x]
    elif isinstance(x, core.NeuronList):
        pers = [persistence_points(n, **kwargs) for n in x]
    elif isinstance(x, list):
        if not all([isinstance(l, pd.DataFrame) for l in x]):
            raise ValueError('Expected lists to contain only DataFrames')
        pers = x
    else:
        raise TypeError('Unable to work extract persistence vectors from data '
                        f'of type "{x}"')

    # Get the max distance
    max_pdist = max([p.birth.max() for p in pers])
    samples = np.linspace(0, max_pdist * 1.05, samples)

    # Now get a persistence vector
    vectors = []
    for p in pers:
        weights = p.death.values - p.birth.values
        if threshold:
            p = p.loc[weights >= threshold]
            weights = weights[weights >= threshold]

        # For each persistence generate a weighted Gaussian kernel
        kernel = gaussian_kde(p.birth.values,
                              weights=weights,
                              bw_method=bw)

        # And sample probabilities at the sample points
        vectors.append(kernel(samples))
    vectors = np.array(vectors)

    if center:
        # Shift each vector such that the highest value lies in the center.
        # Note that we are "rolling" the array which means that elements that
        # drop off to the right are reintroduced on the left
        for i in range(len(vectors)):
            vectors[i] = np.roll(vectors[i],
                                 -np.argmax(vectors[i]) + len(samples) // 2)

    return vectors, samples


def persistence_diagram(pers, ax=None, **kwargs):
    """Plot a persistence diagram.

    Parameters
    ----------
    pers :      pd.DataFrame
                Persistent points from [`navis.persistence_points`][].
    ax :        matplotlib ax, optional
                Ax to plot on.
    **kwargs
                Keyword arguments are passed to `LineCollection`.

    Returns
    -------
    ax :        matplotlib ax

    """
    if not isinstance(pers, pd.DataFrame):
        raise TypeError(f'Expected DataFrame, got "{type(pers)}"')

    if not ax:
        fig, ax = plt.subplots()

    segs = []
    for i, (b, d) in enumerate(zip(pers.birth.values, pers.death.values)):
        segs.append([[b, i], [d, i]])
    lc = LineCollection(segs, **kwargs)
    ax.add_collection(lc)

    ax.set_xlim(-5, pers.death.max())
    ax.set_ylim(-5, pers.shape[0])

    ax.set_ylabel('segments')
    ax.set_xlabel('time')

    return ax


def persistence_vector_plot(x, normalize=True, ax=None,
                            persistence_kwargs={}, vector_kwargs={}):
    """Plot persistence vectors.

    Parameters
    ----------
    x :         TreeNeuron | MeshNeuron | NeuronList
                Neuron(s) to calculate persistence points for. For MeshNeurons,
                we will use the skeleton produced by/associated with its
                `.skeleton` property.

    Returns
    -------
    ax

    """
    if not isinstance(x, core.NeuronList):
        x = core.NeuronList(x)

    # Get persistence points for each skeleton
    pers = persistence_points(x, **persistence_kwargs)

    # Get the vectors
    vectors, samples = persistence_vectors(pers, **vector_kwargs)

    # Normalizing the vectors will produce more useful distances
    if normalize:
        vectors = vectors / vectors.max(axis=1).reshape(-1, 1)
    else:
        vectors = vectors / vectors.max()

    if not ax:
        fig, ax = plt.subplots()

    for n, v in zip(x, vectors):
        ax.plot(samples, v, label=n.label)

    return ax



#!/usr/bin/env python3
# munger.py
# script to process all of the images in the specified directory tree
# 1) Run affine transform
# 2) Run Warp transform
# 3) Reformat

import argparse
import os
import sys
import signal
import gzip
from pathlib import Path

version = "1.21"
is_win = os.name == 'nt'

# Autoflush stdout
sys.stdout.reconfigure(line_buffering=True)

# Global variable to handle QUIT signal
quit_next_image = False

def interrupt(signal, frame):
    global quit_next_image
    quit_next_image = True

signal.signal(signal.SIGQUIT, interrupt)

def init():
    parser = argparse.ArgumentParser(description='Process images in a directory tree.')
    parser.add_argument('-E', type=str, default="1e-1", help='Energy weight')
    parser.add_argument('-X', type=str, default="16", help='Exploration')
    parser.add_argument('-M', type=str, default="nmi", help='Metric')
    parser.add_argument('-C', type=str, default="4", help='Coarsest')
    parser.add_argument('-G', type=str, default="40", help='Grid spacing')
    parser.add_argument('-R', type=str, default="3", help='Refine')
    parser.add_argument('-J', type=str, default="0", help='Jacobian')
    parser.add_argument('-o', type=str, default="nrrd", help='Output type')
    parser.add_argument('-b', type=str, help='Binary directory')
    parser.add_argument('-s', type=str, required=True, help='Reference image')
    parser.add_argument('-v', action='store_true', help='Verbose output')
    parser.add_argument('-T', type=str, default="auto", help='Threads')
    parser.add_argument('-c', type=str, default="01", help='Registration channels')
    parser.add_argument('-r', type=str, default="01", help='Reformat channels')
    parser.add_argument('-l', type=str, default="f", help='Reformat levels')
    parser.add_argument('-f', type=str, default="01", help='Reference image channel')
    parser.add_argument('-d', type=str, help='Output directory')
    parser.add_argument('-k', type=str, help='Lock message')
    parser.add_argument('-x', type=str, default="never", help='Delete input image')
    parser.add_argument('-m', type=int, default=8760, help='Max time in hours')
    parser.add_argument('-p', type=str, help='Script file')
    parser.add_argument('input_files', nargs='+', help='Input files or directories')
    return parser.parse_args()

def arrayidx(value, options):
    try:
        return options.index(value)
    except ValueError:
        return -1

def main():
    args = init()

    energyweight = args.E
    exploration = args.X
    metric = args.M
    coarsest = args.C
    gridspacing = args.G
    refine = args.R
    jacobian = args.J
    output_type = args.o

    metric_options = ["nmi", "mi", "cr", "msd", "ncc"]
    metric_index = arrayidx(metric, metric_options)
    if metric_index == -1:
        sys.exit(f"Unrecognized metric {metric}")

    warp_suffix = f"warp_m{metric_index}g{gridspacing}c{coarsest}e{energyweight}x{exploration}r{refine}"
    icweight = args.I if args.I else "0"

    hostname = os.uname().nodename
    print(f"hostname = {hostname}")

    if is_win:
        if args.v:
            os.system('echo User path is %PATH%')
    else:
        if args.v:
            os.system('echo User path is $PATH')

    threads = args.T
    reference_image = args.s
    if args.v:
        print(f"Reference brain is {reference_image}")

    if not os.path.isfile(reference_image):
        sys.exit(f"Unable to read reference brain {reference_image}")

    reference_stem = Path(reference_image).stem
    reference_stem = reference_stem.replace('_warp', '-warp').replace('_9dof', '-9dof')
    reference_stem = reference_stem.split('.')[0].split('_')[0]
    if args.v:
        print(f"Reference brain stem is {reference_stem}")

    bin_dir = args.b if args.b else "@CMTK_BINARY_DIR_CONFIG@"
    if not os.path.isdir(bin_dir):
        sys.exit(f"Can't access binary directory {bin_dir}")

    warp_command = os.path.join(bin_dir, "warp")
    aff_command = os.path.join(bin_dir, "registration")
    initial_aff_command = os.path.join(bin_dir, "make_initial_affine")
    landmarks_aff_command = os.path.join(bin_dir, "align_landmarks")
    reformat_command = os.path.join(bin_dir, "reformatx")

    reg_channels = args.c
    reformat_channels = args.r
    reformat_levels = args.l
    reference_image_channel = args.f

    reg_root = "Registration"
    image_root = "images"
    reformat_root = "reformatted"

    if args.d:
        dir = args.d
        if dir.startswith('.'):
            reg_root += args.d
            reformat_root += args.d
        else:
            reg_root = args.d

    lockmessage = args.k if args.k else f"{hostname}:{os.getpgrp()}" if not is_win else ""
    print(f"JOB ID = {lockmessage}")

    delete_input_image = args.x
    affine_total = 0
    initial_affine_total = 0
    affine_total_failed = 0
    initial_affine_total_failed = 0
    warp_total = 0
    reformat_total = 0
    maxtime = args.m * 3600
    starttime = time.time()
    if args.v:
        print(f"Start time is: {starttime} seconds")

    found = {}
    if args.p:
        if args.v:
            print(f"Generating script file {args.p}")
        try:
            script_file = open(args.p, 'w')
        except IOError:
            sys.exit(f"Unable to open script file {args.p} for writing")

    root_dir = os.getcwd()
    if is_win:
        root_dir = os.path.abspath(os.sep)
    print(f"Root directory is {root_dir}")

    nargs = len(args.input_files)
    if nargs < 1:
        sys.exit("Usage: munger.py [options] input_files")

    if args.v:
        print(f"There are {nargs} arguments")

    for input_file_spec in args.input_files:
        # Process each input file or directory
        pass

if __name__ == "__main__":
    main()


#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

"""Interface with Allen cell type atlas: https://celltypes.brain-map.org/."""

from textwrap import dedent

try:
    import allensdk
    from allensdk.core.cell_types_cache import CellTypesCache
except ModuleNotFoundError as e:
    msg = dedent("""
          allensdk library not found. Please install using pip:

                pip install allensdk --no-deps

          """)
    raise ModuleNotFoundError(msg) from e
except BaseException:
    raise

import numpy as np
import pandas as pd

from .. import config, utils
from ..core import TreeNeuron, NeuronList

logger = config.get_logger(__name__)
dataset = None

DTYPES = {
    'node_id': np.int32,
    'parent_id': np.int32,
    'compartment': 'category',
    'x': np.float32,
    'y': np.float32,
    'z': np.float32,
    'radius': np.float32,
}
COMPS = {
    1: 'soma',
    2: 'axon',
    3: 'dendrites',
    4: 'apical dendrites'
}
SWC_FILE_TYPE = '3DNeuronReconstruction'

ctc = CellTypesCache()


__all__ = ['fetch_neurons']


def fetch_neurons(ids):
    """Fetch given neurons.

    Parameters
    ----------
    ids :   int | iterable
            IDs of the neurons to fetch skeletons for.

    Returns
    -------
    navis.NeuronList

    """
    ids = utils.make_iterable(ids, force_type=int)

    neurons = []
    for id in config.tqdm(ids,
                          desc='Fetching',
                          disable=config.pbar_hide,
                          leave=config.pbar_leave
                          ):
        morphology = ctc.get_reconstruction(id)
        neurons.append(_parse_morphology(morphology))
        neurons[-1].id = id

    return NeuronList(neurons)


def _parse_morphology(morphology):
    """Convert allensdk morphology to TreeNeuron."""
    assert isinstance(morphology, allensdk.core.swc.Morphology)

    nodes = []
    for n in morphology.compartment_list:
        nodes.append([n[k] for k in ['id', 'x', 'y', 'z', 'radius', 'parent', 'type']])
    nodes = pd.DataFrame(nodes, columns=['node_id', 'x', 'y', 'z', 'radius', 'parent_id', 'compartment'])
    nodes['compartment'] = nodes.compartment.map(COMPS).fillna('na')
    nodes = nodes.astype(DTYPES)

    # I'm guessing these are all in microns
    n = TreeNeuron(nodes, units='1 um')

    if getattr(morphology, 'soma', None):
        n.soma = morphology.soma['id']

    return n



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#    GNU General Public License for more details.

"""Interface with MICrONS datasets: https://www.microns-explorer.org/."""

from .. import config
from functools import lru_cache
from textwrap import dedent
from . import cave_utils

err_msg = dedent("""
      Failed to import `caveclient` library. Please install using pip:

            pip install caveclient -U

      """)

try:
    from caveclient import CAVEclient
    import cloudvolume as cv
except ModuleNotFoundError:
    config.logger.error(err_msg)
    CAVEclient = None
    cv = None
except BaseException:
    raise

NUCLEUS_TABLE = "nucleus_detection_v0"


@lru_cache(None)
def _translate_datastack(datastack):
    """Translate datastack to source."""
    ds = get_datastacks(microns_only=False)

    if datastack in ds:
        return datastack
    elif datastack in ("cortex65", "minnie65"):
        # Find the latest cortex65 datastack
        # "minnie65_public" is apparently the prefered datastack
        if "minnie65_public" in ds:
            return "minnie65_public"
        # If for some reason that stack is not available, just take the latest
        return sorted([d for d in ds if "minnie65_public" in d and "sandbox" not in d])[
            -1
        ]
    elif datastack in ("cortex35", "minnie35"):
        # Find the latest cortex35 datastack
        return sorted([d for d in ds if "minnie35_public" in d and "sandbox" not in d])[
            -1
        ]
    elif datastack == "layer 2/3":
        # The "pinky_sandbox" seems to be the latest layer 2/3 datastack
        return sorted([d for d in ds if "pinky" in d])[-1]
    raise ValueError(f"Datastack '{datastack}' not found.")


@lru_cache(None)
def get_datastacks(microns_only=True):
    """Get available datastacks.

    Parameters
    ----------
    microns_only : bool
        If True, only return MICrONS datastacks.

    Returns
    -------
    list
        List of available datastacks.

    """
    if not CAVEclient:
        raise ModuleNotFoundError(err_msg)

    stacks = CAVEclient().info.get_datastacks()

    if microns_only:
        stacks = [s for s in stacks if "minnie" in s or "pinky" in s]
    return stacks


def get_cave_client(datastack="cortex65"):
    """Get caveclient for given datastack.

    Parameters
    ----------
    datastack :     "cortex65" | "cortex35" | "layer 2/3" | str
                    Which dataset to query. "cortex65", "cortex35" and "layer 2/3"
                    are internally mapped to the corresponding sources: for example,
                    "minnie65_public" for "cortex65"

    """
    if not CAVEclient:
        raise ModuleNotFoundError(err_msg)

    # Try mapping, else pass-through
    datastack = _translate_datastack(datastack)
    client = cave_utils.get_cave_client(datastack)
    client.materialize.nucleus_table = NUCLEUS_TABLE
    return client


@lru_cache(None)
def list_annotation_tables(datastack="cortex65"):
    """Get available annotation tables for given datastack.

    Parameters
    ----------
    datastack :     "cortex65" | "cortex35" | "layer 2/3" | str
                    Which dataset to query. "cortex65", "cortex35" and "layer 2/3"
                    are internally mapped to the corresponding sources: for example,
                    "minnie65_public" for "cortex65"

    Returns
    -------
    list
        List of available annotation tables.

    """
    return get_cave_client(datastack).materialize.get_tables()


def fetch_neurons(
    x,
    *,
    lod=2,
    with_synapses=True,
    datastack="cortex65",
    materialization="auto",
    parallel=True,
    max_threads=4,
    **kwargs,
):
    """Fetch neuron meshes.

    Notes
    -----
    Synapses will be attached to the closest vertex on the mesh.

    Parameters
    ----------
    x :             str | int | list-like
                    Segment ID(s). Multiple Ids can be provided as list-like.
    lod :           int
                    Level of detail. Higher ``lod`` = coarser. This parameter
                    is ignored if the data source does not support multi-level
                    meshes.
    with_synapses : bool, optional
                    If True will also attach synapses as ``.connectors``.
    datastack :     "cortex65" | "cortex35" | "layer 2/3" | str
                    Which dataset to query. "cortex65", "cortex35" and "layer 2/3"
                    are internally mapped to the corresponding sources: for example,
                    "minnie65_public" for "cortex65"
    materialization : "auto" | int
                    Which materialization version to use to look up somas and synapses
                    (if applicable). If "auto" (default) will try to find the most
                    recent version that contains the given root IDs. If an
                    integer is provided will use that version.
    parallel :      bool
                    If True, will use parallel threads to fetch data.
    max_threads :   int
                    Max number of parallel threads to use.
    **kwargs
                    Keyword arguments are passed through to the initialization
                    of the ``navis.MeshNeurons``.

    Returns
    -------
    navis.Neuronlist
                    Containing :class:`navis.MeshNeuron`.

    """
    client = get_cave_client(_translate_datastack(datastack))
    return cave_utils.fetch_neurons(
        x,
        lod=lod,
        with_synapses=with_synapses,
        client=client,
        parallel=parallel,
        max_threads=max_threads,
        materialization=materialization,
        **kwargs,
    )


def get_voxels(x, mip=0, bounds=None, datastack="cortex65"):
    """Fetch voxels making a up given root ID.

    Parameters
    ----------
    x :             int
                    A single root ID.
    mip :           int
                    Scale at which to fetch voxels. Lower = higher resolution.
    bounds :        list, optional
                    Bounding box [xmin, xmax, ymin, ymax, zmin, zmax] in voxel
                    space. For example, the voxel resolution for mip 0
                    segmentation is 8 x 8 x 40 nm.
    datastack :     "cortex65" | "cortex35" | "layer 2/3" | str
                    Which dataset to query. "cortex65", "cortex35" and "layer 2/3"
                    are internally mapped to the corresponding sources: for example,
                    "minnie65_public" for "cortex65"

    Returns
    -------
    voxels :        (N, 3) np.ndarray
                    In voxel space according to `mip`.

    """
    return cave_utils.get_voxels(
        x, mip=mip, bounds=bounds, client=get_cave_client(datastack)
    )



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import uuid
import warnings

import numpy as np
import pandas as pd
import trimesh as tm

import plotly.graph_objs as go

from scipy import ndimage

from ..colors import vertex_colors, eval_color
from ..plot_utils import segments_to_coords
from ... import core, utils, config, conversion

logger = config.get_logger(__name__)

__all__ = [
    "neuron2plotly",
    "scatter2plotly",
    "dotprops2plotly",
    "volume2plotly",
    "layout2plotly",
]

# Generate sphere for somas
BASE_SPHERE = tm.primitives.Sphere(radius=1, center=(0, 0, 0), subdivisions=2)


def neuron2plotly(x, colormap, settings):
    """Convert neurons to plotly objects."""
    if isinstance(x, core.BaseNeuron):
        x = core.NeuronList(x)
    elif not isinstance(x, core.NeuronList):
        raise TypeError('Unable to process data of type "{}"'.format(type(x)))

    if not isinstance(settings.color_by, type(None)):
        if not settings.palette:
            raise ValueError(
                'Must provide `palette` (e.g. "viridis") argument '
                "if using `color_by`"
            )

        colormap = vertex_colors(
            x,
            by=settings.color_by,
            alpha=settings.alpha,
            use_alpha=False,
            palette=settings.palette,
            vmin=settings.vmin,
            vmax=settings.vmax,
            na="raise",
            color_range=255,
        )

    if settings.shade_by is not None:
        alphamap = vertex_colors(
            x,
            by=settings.shade_by,
            use_alpha=True,
            palette="viridis",  # palette is irrelevant here
            vmin=settings.smin,
            vmax=settings.smax,
            na="raise",
            color_range=255,
        )

        new_colormap = []
        for c, a in zip(colormap, alphamap):
            if not (isinstance(c, np.ndarray) and c.ndim == 2):
                c = np.tile(c, (a.shape[0], 1))

            if c.dtype not in (np.float16, np.float32, np.float64):
                c = c.astype(np.float16)

            if c.shape[1] == 4:
                c[:, 3] = a[:, 3]
            else:
                c = np.insert(c, 3, a[:, 3], axis=1)

            new_colormap.append(c)
        colormap = new_colormap

    cn_lay = config.default_connector_colors.copy()
    cn_lay.update(settings.cn_layout)

    trace_data = []
    _radius_warned = False
    for i, neuron in enumerate(x):
        name = str(getattr(neuron, "name", neuron.id))
        color = colormap[i]

        try:
            # Try converting this neuron's ID
            neuron_id = str(neuron.id)
        except BaseException:
            # If that doesn't work generate a new ID
            neuron_id = str(str(uuid.uuid1()))

        showlegend = True
        label = neuron.label
        if (
            isinstance(settings.legend_group, dict)
            and neuron.id in settings.legend_group
        ):
            # Check if this the first entry for this legendgroup
            legendgroup = settings.legend_group[neuron.id]
            for d in trace_data:
                # If it is not the first entry, hide it
                if getattr(d, "legendgroup", None) == settings.legend_group:
                    showlegend = False
                    break
        elif isinstance(settings.legend_group, str):
            legendgroup = settings.legend_group
        else:
            legendgroup = neuron_id

        if isinstance(neuron, core.TreeNeuron) and settings.radius == "auto":
            # Number of nodes with radii
            n_radii = (neuron.nodes.get("radius", pd.Series([])).fillna(0) > 0).sum()
            # If less than 30% of nodes have a radius, we will fall back to lines
            if n_radii / neuron.nodes.shape[0] < 0.3:
                settings.radius = False

        if isinstance(neuron, core.TreeNeuron) and settings.radius:
            # Warn once if more than 5% of nodes have missing radii
            if not _radius_warned:
                if (
                    (neuron.nodes.radius.fillna(0).values <= 0).sum() / neuron.n_nodes
                ) > 0.05:
                    logger.warning(
                        "Some skeleton nodes have radius <= 0. This may lead to "
                        "rendering artifacts. Set `radius=False` to plot skeletons "
                        "as single-width lines instead."
                    )
                    _radius_warned = True

            _neuron = conversion.tree2meshneuron(
                neuron,
                warn_missing_radii=False,
                radius_scale_factor=settings.get("linewidth", 1),
            )
            _neuron.connectors = neuron.connectors
            neuron = _neuron

            # See if we need to map colors to vertices
            if isinstance(color, np.ndarray) and color.ndim == 2:
                color = color[neuron.vertex_map]

        if not settings.connectors_only:
            if isinstance(neuron, core.TreeNeuron):
                trace_data += skeleton2plotly(
                    neuron,
                    label=label,
                    legendgroup=legendgroup,
                    showlegend=showlegend,
                    color=color,
                    settings=settings,
                )
            elif isinstance(neuron, core.MeshNeuron):
                trace_data += mesh2plotly(
                    neuron,
                    label=label,
                    legendgroup=legendgroup,
                    showlegend=showlegend,
                    color=color,
                    settings=settings,
                )
            elif isinstance(neuron, core.VoxelNeuron):
                trace_data += voxel2plotly(
                    neuron,
                    label=label,
                    legendgroup=legendgroup,
                    showlegend=showlegend,
                    color=color,
                    settings=settings,
                )
            elif isinstance(neuron, core.Dotprops):
                trace_data += dotprops2plotly(
                    neuron,
                    label=label,
                    legendgroup=legendgroup,
                    showlegend=showlegend,
                    color=color,
                    settings=settings,
                )
            else:
                raise TypeError(f'Unable to plot neurons of type "{type(neuron)}"')

        # Add connectors
        if (settings.connectors or settings.connectors_only) and neuron.has_connectors:
            if isinstance(settings.connectors, (list, np.ndarray, tuple)):
                connectors = neuron.connectors[
                    neuron.connectors.type.isin(settings.connectors)
                ]
            elif settings.connectors == "pre":
                connectors = neuron.presynapses
            elif settings.connectors == "post":
                connectors = neuron.postsynapses
            elif isinstance(settings.connectors, str):
                connectors = neuron.connectors[
                    neuron.connectors.type == settings.connectors
                ]
            else:
                connectors = neuron.connectors

            for j, this_cn in connectors.groupby("type"):
                if isinstance(settings.cn_colors, dict):
                    c = settings.cn_colors.get(
                        j, cn_lay.get(j, {"color": (10, 10, 10)})["color"]
                    )
                elif settings.cn_colors == "neuron":
                    c = color
                elif settings.cn_colors is not None:
                    c = settings.cn_colors
                else:
                    c = cn_lay.get(j, {"color": (10, 10, 10)})["color"]

                c = eval_color(c, color_range=255)

                if cn_lay["display"] == "circles" or isinstance(
                    neuron, core.MeshNeuron
                ):
                    trace_data.append(
                        go.Scatter3d(
                            x=this_cn.x.values,
                            y=this_cn.y.values,
                            z=this_cn.z.values,
                            mode="markers",
                            marker=dict(
                                color=f"rgb{c}",
                                size=settings.cn_size
                                if settings.cn_size
                                else cn_lay["size"],
                            ),
                            name=f'{cn_lay.get(j, {"name": "connector"})["name"]} of {name}',
                            showlegend=False,
                            legendgroup=legendgroup,
                            hoverinfo="none",
                        )
                    )
                elif cn_lay["display"] == "lines":
                    # Find associated treenodes
                    tn = neuron.nodes.set_index("node_id").loc[this_cn.node_id.values]
                    x_coords = [
                        n
                        for sublist in zip(
                            this_cn.x.values, tn.x.values, [None] * this_cn.shape[0]
                        )
                        for n in sublist
                    ]
                    y_coords = [
                        n
                        for sublist in zip(
                            this_cn.y.values, tn.y.values, [None] * this_cn.shape[0]
                        )
                        for n in sublist
                    ]
                    z_coords = [
                        n
                        for sublist in zip(
                            this_cn.z.values, tn.z.values, [None] * this_cn.shape[0]
                        )
                        for n in sublist
                    ]

                    trace_data.append(
                        go.Scatter3d(
                            x=x_coords,
                            y=y_coords,
                            z=z_coords,
                            mode="lines",
                            line=dict(color="rgb%s" % str(c), width=5),
                            name=f'{cn_lay.get(j, {"name": "connector"})["name"]} of {name}',
                            showlegend=False,
                            legendgroup=legendgroup,
                            hoverinfo="none",
                        )
                    )
                else:
                    raise ValueError(
                        f'Unknown display type for connectors "{cn_lay["display"]}"'
                    )

    return trace_data


def mesh2plotly(neuron, legendgroup, showlegend, label, color, settings):
    """Convert MeshNeuron to plotly object."""
    # Skip empty neurons
    if neuron.n_vertices == 0:
        return []

    if isinstance(color, np.ndarray) and color.ndim == 2:
        if len(color) == len(neuron.vertices):
            # For some reason single colors are 0-255 but face/vertex colors
            # have to be 0-1
            color_kwargs = dict(
                vertexcolor=color / [255, 255, 255, 1][: color.shape[1]]
            )
        elif len(color) == len(neuron.faces):
            color_kwargs = dict(facecolor=color / [255, 255, 255, 1][: color.shape[1]])
        else:
            color_kwargs = dict(color=color)
    else:
        try:
            if len(color) == 3:
                c = "rgb{}".format(color)
            elif len(color) == 4:
                c = "rgba{}".format(color)
        except BaseException:
            c = "rgb(10,10,10)"
        color_kwargs = dict(color=c)

    if settings.hover_name:
        hoverinfo = "text"
        hovertext = neuron.label
    else:
        hoverinfo = "none"
        hovertext = " "

    trace_data = [
        go.Mesh3d(
            x=neuron.vertices[:, 0],
            y=neuron.vertices[:, 1],
            z=neuron.vertices[:, 2],
            i=neuron.faces[:, 0],
            j=neuron.faces[:, 1],
            k=neuron.faces[:, 2],
            name=label,
            legendgroup=legendgroup,
            legendgrouptitle_text=legendgroup,
            showlegend=showlegend,
            hovertext=hovertext,
            hoverinfo=hoverinfo,
            **color_kwargs,
        )
    ]

    return trace_data


def voxel2plotly(
    neuron, legendgroup, showlegend, label, color, settings, as_scatter=True
):
    """Convert VoxelNeuron to plotly object.

    Turns out that plotly is horrendous for plotting voxel data (Volumes):
    anything more than a few thousand voxels (e.g. 40x40x40) and the html
    encoding and loading the plot takes ages. Unfortunately, the same happens
    with Isosurfaces.

    I'm adding an implementation here but until plotly gets MUCH better at this,
    there is really no point. For now, we will fallback to plotting the
    voxels as scatter plots using the top 10k voxels sorted by brightness.

    """
    # Skip empty neurons
    if min(neuron.shape) == 0:
        return []

    try:
        if len(color) == 3:
            c = "rgb{}".format(color)
        elif len(color) == 4:
            c = "rgba{}".format(color)
    except BaseException:
        c = "rgb(10,10,10)"

    if settings.hover_name:
        hoverinfo = "text"
        hovertext = neuron.label
    else:
        hoverinfo = "none"
        hovertext = " "

    if not as_scatter:
        # Downsample heavily
        ds = ndimage.zoom(neuron.grid, 0.2, order=1)

        # Generate X, Y, Z, coordinates for values in grid
        X, Y, Z = np.meshgrid(
            range(ds.shape[0]), range(ds.shape[1]), range(ds.shape[2]), indexing="ij"
        )

        # Flatten and scale coordinates
        X = X.flatten() * neuron.units_xyz[0] + neuron.offset[0]
        Y = Y.flatten() * neuron.units_xyz[1] + neuron.offset[1]
        Z = Z.flatten() * neuron.units_xyz[2] + neuron.offset[2]

        # Flatten and normalize values
        values = ds.flatten() / ds.max()

        trace_data = [
            go.Isosurface(
                x=X,
                y=Y,
                z=Z,
                value=values,
                isomin=0.001,
                isomax=1,
                opacity=0.1,
                surface_count=21,
            )
        ]
    else:
        voxels, values = neuron.voxels, neuron.values

        # Sort by brightness
        srt = np.argsort(values)

        # Take the top 100k voxels
        values = values[srt[-100000:]]
        voxels = voxels[srt[-100000:]]

        # Scale and offset voxels
        voxels = voxels * neuron.units_xyz.magnitude + neuron.offset

        with warnings.catch_warnings():
            trace_data = [
                go.Scatter3d(
                    x=voxels[:, 0],
                    y=voxels[:, 1],
                    z=voxels[:, 2],
                    mode="markers",
                    marker=dict(
                        color=values, size=4, colorscale="viridis", opacity=0.1
                    ),
                    name=label,
                    legendgroup=legendgroup,
                    legendgrouptitle_text=legendgroup,
                    showlegend=showlegend,
                    hovertext=hovertext,
                    hoverinfo=hoverinfo,
                )
            ]

    return trace_data


def skeleton2plotly(neuron, legendgroup, showlegend, label, color, settings):
    """Convert skeleton (i.e. TreeNeuron) to plotly line plot."""
    if not hasattr(neuron, "nodes") or neuron.nodes.empty:
        logger.warning(f"Skipping TreeNeuron w/o nodes: {neuron.label}")
        return []
    elif neuron.nodes.shape[0] == 1:
        logger.warning(f"Skipping single-node TreeNeuron: {neuron.label}")
        return []

    coords = segments_to_coords(neuron)

    # For some reason, plotly seems to ignore the alpha channel when given an RGBA color
    # (the color still changes somewhat but the line doesn't turn transparent)
    # Instead, we have to set the `opacity` property of the whole scatter object
    # - we will adjust opacity further down if according to the color
    opacity = 1

    # We have to add (None, None, None) to the end of each segment to
    # make that line discontinuous
    coords = np.vstack([np.append(t, [[None] * 3], axis=0) for t in coords])

    if isinstance(color, np.ndarray) and color.ndim == 2:
        # Change colors to rgb/a strings
        if color.shape[1] == 4:
            c = [f"rgba({c[0]},{c[1]},{c[2]},{c[3]:.3f})" for c in color]
        else:
            c = [f"rgb({c[0]},{c[1]},{c[2]})" for c in color]

        # Next we have to make colors match the segments in `coords`
        c = np.asarray(c)
        ix = dict(zip(neuron.nodes.node_id.values, np.arange(neuron.n_nodes)))
        c = [
            col
            for s in neuron.segments
            for col in np.append(c[[ix[n] for n in s]], "rgb(0,0,0)")
        ]

    else:
        if len(color) == 4:
            opacity = color[3]

        c = f"rgb({color[0]},{color[1]},{color[2]})"

    if settings.hover_id:
        hoverinfo = "text"
        hovertext = [str(i) for seg in neuron.segments for i in seg + [None]]
    elif settings.hover_name:
        hoverinfo = "text"
        hovertext = neuron.label
    else:
        hoverinfo = "none"
        hovertext = " "

    # Options for linestyle: "solid", "dot", "dash", "longdash", "dashdot", or "longdashdot"
    # Translate `linestyle` setting to plotly's `dash` setting
    dash = {"-": "solid", "--": "dash", "-.": "dashdot", ":": "dot"}.get(
        settings.linestyle, settings.linestyle
    )

    trace_data = [
        go.Scatter3d(
            x=coords[:, 0],
            y=coords[:, 1],
            z=coords[:, 2],
            opacity=opacity,
            mode="lines",
            line=dict(color=c, width=settings.get('linewidth', 3), dash=dash),
            name=label,
            legendgroup=legendgroup,
            legendgrouptitle_text=legendgroup,
            showlegend=showlegend,
            hoverinfo=hoverinfo,
            hovertext=hovertext,
        )
    ]

    # Add soma(s):
    soma = utils.make_iterable(neuron.soma)
    if settings.soma:
        # If soma detection is messed up we might end up producing
        # hundrets of soma which will freeze the session
        if len(soma) >= 10:
            logger.warning(
                f"Neuron {neuron.id} appears to have {len(soma)} "
                "somas. That does not look right - will ignore "
                "them for plotting."
            )
        else:
            for s in soma:
                # Skip `None` somas
                if isinstance(s, type(None)):
                    continue

                # If we have colors for every vertex, we need to find the
                # color that corresponds to this root (or it's parent to be
                # precise)
                if isinstance(c, list):
                    s_ix = np.where(neuron.nodes.node_id == s)[0][0]
                    soma_color = c[s_ix]
                else:
                    soma_color = c

                n = neuron.nodes.set_index("node_id").loc[s]
                r = (
                    getattr(n, neuron.soma_radius)
                    if isinstance(neuron.soma_radius, str)
                    else neuron.soma_radius
                )

                trace_data += [
                    go.Mesh3d(
                        x=BASE_SPHERE.vertices[:, 0] * r + n.x,
                        y=BASE_SPHERE.vertices[:, 1] * r + n.y,
                        z=BASE_SPHERE.vertices[:, 2] * r + n.z,
                        i=BASE_SPHERE.faces[:, 0],
                        j=BASE_SPHERE.faces[:, 1],
                        k=BASE_SPHERE.faces[:, 2],
                        legendgroup=legendgroup,
                        showlegend=False,
                        hoverinfo="name",
                        color=soma_color,
                    )
                ]

    return trace_data


def scatter2plotly(x, **scatter_kws):
    """Convert DataFrame with x,y,z columns to plotly scatter plot."""
    c = eval_color(
        scatter_kws.get("color", scatter_kws.get("c", (100, 100, 100))), color_range=255
    )
    s = scatter_kws.get("size", scatter_kws.get("s", 2))
    name = scatter_kws.get("name", None)

    trace_data = []
    for scatter in x:
        if isinstance(scatter, pd.DataFrame):
            if not all([c in scatter.columns for c in ["x", "y", "z"]]):
                raise ValueError("DataFrame must have x, y and z columns")
            scatter = scatter[["x", "y", "z"]].values

        if not isinstance(scatter, np.ndarray):
            scatter = np.array(scatter)

        trace_data.append(
            go.Scatter3d(
                x=scatter[:, 0],
                y=scatter[:, 1],
                z=scatter[:, 2],
                mode=scatter_kws.get("mode", "markers"),
                marker=dict(
                    color="rgb%s" % str(c),
                    size=s,
                    opacity=scatter_kws.get("opacity", 1),
                ),
                name=name,
                showlegend=True,
                hoverinfo="none",
            )
        )
    return trace_data


def lines2plotly(x, **kwargs):
    """Convert DataFrame with x, y, z, x1, y1, z1 columns to line plots."""
    name = kwargs.get("name", None)
    c = kwargs.get("color", (10, 10, 10))

    x_coords = [
        n
        for sublist in zip(x.x.values, x.x1.values, [None] * x.shape[0])
        for n in sublist
    ]
    y_coords = [
        n
        for sublist in zip(x.y.values, x.y1.values, [None] * x.shape[0])
        for n in sublist
    ]
    z_coords = [
        n
        for sublist in zip(x.z.values, x.z1.values, [None] * x.shape[0])
        for n in sublist
    ]

    trace_data = []
    trace_data.append(
        go.Scatter3d(
            x=x_coords,
            y=y_coords,
            z=z_coords,
            mode="lines",
            line=dict(color=f"rgb{str(c)}", width=5),
            name=name,
            showlegend=True,
            hoverinfo="none",
        )
    )

    return trace_data


def dotprops2plotly(x, legendgroup, showlegend, label, color, settings):
    """Convert Dotprops to plotly graph object."""
    return skeleton2plotly(
        x.to_skeleton(scale_vec=settings.dps_scale_vec),
        legendgroup,
        showlegend,
        label,
        color,
        settings,
    )


def volume2plotly(x, colormap, settings):
    """Convert Volumes to plotly objects."""
    trace_data = []
    for i, v in enumerate(x):
        # Skip empty data
        if isinstance(v.vertices, np.ndarray):
            if not v.vertices.any():
                continue
        elif not v.vertices:
            continue

        name = getattr(v, "name", None)

        c = colormap[i]
        if len(c) == 3:
            c = (c[0], c[1], c[2], 0.5)

        rgba_str = f"rgba({c[0]:.0f},{c[1]:.0f},{c[2]:.0f},{c[3]:.1f})"
        trace_data.append(
            go.Mesh3d(
                x=v.vertices[:, 0],
                y=v.vertices[:, 1],
                z=v.vertices[:, 2],
                i=v.faces[:, 0],
                j=v.faces[:, 1],
                k=v.faces[:, 2],
                color=rgba_str,
                name=name,
                showlegend=settings.volume_legend,
                hoverinfo="none",
            )
        )

    return trace_data


def layout2plotly(**kwargs):
    """Generate layout for plotly figures."""
    layout = dict(
        width=kwargs.get("width", None),  # these override autosize
        height=kwargs.get("height", 600),  # these override autosize
        autosize=kwargs.get("fig_autosize", True),
        title=kwargs.get("pl_title", None),
        showlegend=kwargs.get("legend", True),
        legend_orientation=kwargs.get("legend_orientation", "v"),
        legend_bgcolor="rgba(0,0,0,0)",
        plot_bgcolor="rgba(0,0,0,0)",
        paper_bgcolor="rgba(0,0,0,0)",
        scene=dict(
            xaxis=dict(
                gridcolor="rgb(255, 255, 255)",
                zerolinecolor="rgb(255, 255, 255)",
                showbackground=False,
                showline=False,
                showgrid=False,
                backgroundcolor="rgb(240, 240, 240)",
            ),
            yaxis=dict(
                gridcolor="rgb(255, 255, 255)",
                zerolinecolor="rgb(255, 255, 255)",
                showbackground=False,
                showline=False,
                showgrid=False,
                backgroundcolor="rgb(240, 240, 240)",
            ),
            zaxis=dict(
                gridcolor="rgb(255, 255, 255)",
                zerolinecolor="rgb(255, 255, 255)",
                showbackground=False,
                showline=False,
                showgrid=False,
                backgroundcolor="rgb(240, 240, 240)",
            ),
            camera=dict(
                up=dict(x=0, y=0, z=1),
                eye=dict(
                    x=-1.7428,
                    y=1.0707,
                    z=0.7100,
                ),
            ),
            aspectratio=dict(x=1, y=1, z=1),
            aspectmode="data",
        ),
    )

    return layout



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import trimesh as tm
import numpy as np

from typing import Union, Optional
from typing_extensions import Literal

from .. import core, config, utils
from .converters import (neuron2voxels, mesh2skeleton, tree2meshneuron,
                         points2skeleton)
from .meshing import voxels2mesh

logger = config.get_logger(__name__)


@utils.map_neuronlist(desc='Skeletonizing', allow_parallel=True)
def skeletonize(x: Union['core.MeshNeuron', 'core.Dotprops', np.ndarray],
                **kwargs):
    """Turn neuron into skeleton.

    Currently, we can only skeletonize meshes, dotprops and point clouds but
    are looking into ways to also do it for `VoxelNeurons`.

    For meshes, this function is a thin-wrapper for `skeletor`. It uses sensible
    defaults for neurons but if you want to fine-tune your skeletons you should
    look into using `skeletor` directly.

    Parameters
    ----------
    x :         MeshNeuron | trimesh.Trimesh | Dotprops
                Mesh(es) to skeletonize. Note that the quality of the results
                very much depends on the mesh, so it might be worth doing some
                pre-processing (see below).
    **kwargs
                Keyword arguments are passed through to the respective
                converters:
                    - meshes: [`navis.conversion.mesh2skeleton`][]
                    - dotprops and point clouds: [`navis.conversion.points2skeleton`][]

    Returns
    -------
    skeleton :  navis.TreeNeuron
                For meshes, this has a `.vertex_map` attribute that maps each
                vertex in the input mesh to a skeleton node ID.

    See Also
    --------
    [`navis.drop_fluff`][]
                Use this if your mesh has lots of tiny free floating bits to
                reduce noise and speed up skeletonization.

    Examples
    --------
    # Skeletonize a mesh
    >>> import navis
    >>> # Get a mesh neuron
    >>> n = navis.example_neurons(1, kind='mesh')
    >>> # Convert to skeleton
    >>> sk = navis.skeletonize(n)
    >>> # Mesh vertex indices to node IDs map
    >>> sk.vertex_map                                           # doctest: +SKIP
    array([938, 990, 990, ...,  39, 234, 234])

    # Skeletonize dotprops (i.e. point-clouds)
    >>> import navis
    >>> # Get a skeleton and turn into dotprops
    >>> dp = navis.make_dotprops(navis.example_neurons(1))
    >>> # Turn back into a skeleton
    >>> sk = navis.skeletonize(dp)

    """
    if isinstance(x, (core.MeshNeuron, tm.Trimesh)):
        return mesh2skeleton(x, **kwargs)
    elif isinstance(x, (core.Dotprops, )):
        sk = points2skeleton(x.points, **kwargs)
        for attr in ('id', 'units', 'name'):
            if hasattr(x, attr):
                setattr(sk, attr, getattr(x, attr))
        return sk
    elif isinstance(x, np.ndarray):
        return points2skeleton(x.points, **kwargs)

    raise TypeError(f'Unable to skeletonize data of type {type(x)}')


@utils.map_neuronlist(desc='Voxelizing', allow_parallel=True)
def voxelize(x: 'core.BaseNeuron',
             pitch: Union[list, tuple, float],
             bounds: Optional[list] = None,
             counts: bool = False,
             vectors: bool = False,
             alphas: bool = False,
             smooth: int = 0) -> 'core.VoxelNeuron':
    """Turn neuron into voxels.

    Parameters
    ----------
    x :             TreeNeuron | MeshNeuron | Dotprops
                    Neuron(s) to voxelize. Uses the neurons' nodes, vertices and
                    points, respectively.
    pitch :         float | iterable thereof
                    Side length(s) voxels. Can be isometric (float) or an
                    iterable of dimensions in (x, y, z).
    bounds :        (3, 2)  or (2, 3) array, optional
                    Boundaries [in units of `x`] for the voxel grid. If not
                    provided, will use `x.bbox`.
    counts :        bool
                    If True, voxel grid will have point counts for values
                    instead of just True/False.
    vectors :       bool
                    If True, will also attach a vector field as `.vectors`
                    property.
    alphas :        bool
                    If True, will also return a grid with alpha values as
                    `.alpha` property.
    smooth :        int
                    If non-zero, will apply a Gaussian filter with `smooth`
                    as `sigma`.

    Returns
    -------
    VoxelNeuron
                    Has the voxel grid as `.grid` and (optionally) `.vectors`
                    and `.alphas` properties. `.grid` data type depends
                    on settings:
                     - default = bool (i.e. True/False)
                     - if `counts=True` = integer
                     - if `smooth=True` = float
                    Empty voxels will have vector (0, 0, 0) and alpha 0. Also
                    note that data tables (e.g. `connectors`) are not carried
                    over from the input neuron.

    Examples
    --------
    >>> import navis
    >>> # Get a skeleton
    >>> n = navis.example_neurons(1)
    >>> # Convert to voxel neuron
    >>> vx = navis.voxelize(n, pitch='5 microns')

    """
    if isinstance(x, (core.TreeNeuron, core.MeshNeuron, core.Dotprops)):
        return neuron2voxels(x,
                             pitch=pitch,
                             bounds=bounds,
                             counts=counts,
                             vectors=vectors,
                             alphas=alphas,
                             smooth=smooth)

    raise TypeError(f'Unable to voxelize data of type {type(x)}')


@utils.map_neuronlist(desc='Meshing', allow_parallel=True)
def mesh(x: Union['core.VoxelNeuron', np.ndarray, 'core.TreeNeuron'],
         **kwargs) -> Union[tm.Trimesh, 'core.MeshNeuron']:
    """Generate mesh from object(s).

    VoxelNeurons or (N, 3) arrays of voxel coordinates will be meshed using
    a marching cubes algorithm. TreeNeurons will be meshed by creating
    cylinders using the radii.

    Parameters
    ----------
    x :             VoxelNeuron | (N, 3) np.array | TreeNeuron
                    Object to mesh. See notes above.
    **kwargs
                    Keyword arguments are passed through to the respective
                    converters: [`navis.conversion.voxels2mesh`][] and
                    [`navis.conversion.tree2meshneuron`][], respectively.

    Returns
    -------
    mesh :          trimesh.Trimesh | MeshNeuron
                    Returns a trimesh or MeshNeuron depending on the input.
                    Data tables (e.g. `connectors`) are not carried over from
                    input neuron.

    """
    if isinstance(x, core.VoxelNeuron) or (isinstance(x, np.ndarray) and x.ndims == 2 and x.shape[1] == 3):
        return voxels2mesh(x, **kwargs)
    elif isinstance(x, core.TreeNeuron):
        return tree2meshneuron(x, **kwargs)

    raise TypeError(f'Unable to create mesh from data of type {type(x)}')



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
"""Module contains functions to plot neurons as flat structures."""

import math
import time

import matplotlib.pyplot as plt
import matplotlib.colors as mcl
import pandas as pd
import numpy as np
import networkx as nx

from typing import Optional, Union, Any, List
from typing_extensions import Literal

from .. import core, config, utils
from ..morpho.mmetrics import parent_dist
from .colors import prepare_connector_cmap

logger = config.get_logger(__name__)

__all__ = ["plot_flat", "print_skeleton"]

_DEFAULTS = dict(
    origin=(0, 0),  # Origin in coordinate system
    start_angle=0,  # Start angle (0 -> to right)
    angle_change=45,  # Angle between branch and its child
    angle_decrease=0,  # Angle decrease with each branch point
    syn_marker_size=0.5,  # Length of orthogonal synapse markers/size of scatter
    switch_dist=1,  # Distance threshold for inverting angle (i.e. flip branch direction)
    syn_linewidth=1.5,  # Line width for connectors
    syn_highlight_color=(1, 0, 0),  # Color for highlighted connectors
    force_nx=False,  # Force using networkx over igraph
    color=(0.1, 0.1, 0.1),  # Color for neurites
)

_NODE_POSITIONS = {}


def plot_flat(
    x,
    layout: Union[
        Literal["subway"],
        Literal["dot"],
        Literal["neato"],
        Literal["fpd"],
        Literal["sfpd"],
        Literal["twopi"],
        Literal["circo"],
    ] = "subway",
    connectors: bool = False,
    highlight_connectors: Optional[List[int]] = None,
    shade_by_length: bool = False,
    normalize_distance: bool = False,
    reroot_soma: bool = False,
    ax: Optional[Any] = None,
    **kwargs,
):
    """Plot neuron as flat diagrams.

    Parameters
    ----------
    x :                     TreeNeuron
                            A single neuron to plot.
    layout :                'subway' | 'dot' | 'neato' | 'fdp' | 'sfpd' | 'twopi' | 'circo'
                            Layout to use. All but 'subway' require graphviz to
                            be installed. For the 'fdp' and 'neato' it is highly
                            recommended to downsample the neuron first.
    connectors :            bool
                            If True (and neuron actually has connectors), will plot
                            connectors.
    highlight_connectors :  list of connector IDs, optional
                            Will highlight these connector IDs.
    ax :                    matplotlib.ax, optional
                            Ax to plot on. Will create new one if not provided.
    shade_by_length :       bool, optional
                            Change shade of branch with length. For layout
                            "subway" only.
    normalize_distance :    bool, optional
                            If True, will normalise all distances to the longest
                            neurite. For layout "subway" only.
    **kwargs
                            Keyword argument passed on to the respective
                            plotting functions.

    Returns
    -------
    ax :                    matplotlib.ax
    pos :                   dict
                            (X, Y) positions for each node: `{node_id: (x, y)}`.


    Examples
    --------
    Plot neuron in "subway" layout:

    >>> import navis
    >>> n = navis.example_neurons(1).convert_units('nm')
    >>> ax, pos = navis.plot_flat(n, layout='subway',
    ...                           figsize=(12, 2),
    ...                           connectors=True)
    >>> _ = ax.set_xlabel('distance [nm]')
    >>> plt.show() # doctest: +SKIP

    Plot neuron in "dot" layout (requires pygraphviz and graphviz):

    >>> # First downsample to speed up processing
    >>> ds = navis.downsample_neuron(n, 10, preserve_nodes='connectors')
    >>> ax, pos = navis.plot_flat(ds, layout='dot', connectors=True) # doctest: +SKIP
    >>> plt.show()                                                   # doctest: +SKIP

    To close all figures (only for doctests)

    >>> plt.close('all')

    See the [plotting intro](../../generated/gallery/1_plotting/tutorial_plotting_00_intro)
    and the [neuron topology tutorial](../../generated/gallery/1_plotting/tutorial_plotting_03_dend)
    for more examples.

    """
    if isinstance(x, core.NeuronList) and len(x) == 1:
        x = x[0]

    utils.eval_param(x, name="x", allowed_types=(core.TreeNeuron,))
    utils.eval_param(
        layout,
        name="layout",
        allowed_values=("subway", "dot", "neato", "fdp", "sfdp", "twopi", "circo"),
    )

    # Work on the copy of the neuron
    x = x.copy()

    # Reroot to soma (if applicable)
    if reroot_soma and x.soma:
        x.reroot(x.soma, inplace=True)

    if layout == "subway":
        return _plot_subway(
            x,
            connectors=connectors,
            highlight_connectors=highlight_connectors,
            shade_by_length=shade_by_length,
            normalize_distance=normalize_distance,
            ax=ax,
            **kwargs,
        )
    else:
        return _plot_force(
            x,
            prog=layout,
            connectors=connectors,
            highlight_connectors=highlight_connectors,
            ax=ax,
            **kwargs,
        )


def _plot_subway(
    x,
    connectors=False,
    highlight_connectors=[],
    shade_by_length=False,
    normalize_distance=False,
    ax=None,
    **kwargs,
):
    """Plot neuron as dendrogram. Preserves distances along branches."""
    DEFAULTS = _DEFAULTS.copy()
    DEFAULTS.update(kwargs)
    if len(x.root) > 1:
        raise ValueError(
            "Unable to plot neuron with multiple roots. Use "
            "`navis.heal_skeleton` to merge the fragments."
        )

    # Change scale of markers if we normalise to max neurite length
    if normalize_distance:
        DEFAULTS["syn_marker_len"] /= 1000
        DEFAULTS["switch_dist"] /= 1000

    if not ax:
        fig, ax = plt.subplots(figsize=kwargs.get("figsize", (10, 10)))
        # Make background transparent (nicer for dark themes)
        fig.patch.set_alpha(0)
        ax.patch.set_alpha(0)

    # For each node get the distance to its root
    if "parent_dist" not in x.nodes.columns:
        x.nodes["parent_dist"] = parent_dist(x, 0)

    # First collect leafs, branches and root
    leaf_nodes = x.leafs.node_id.values
    root_nodes = x.root
    branch_nodes = set(x.branch_points.node_id.values)

    # Use igraph if possible:
    if x.igraph and not DEFAULTS["force_nx"]:
        # Convert node IDs to igraph vertex indices
        leaf_vs = x.igraph.vs.select(node_id_in=leaf_nodes)
        root_vs = x.igraph.vs.select(node_id_in=root_nodes)

        # Now get paths from all tips to the root
        paths = x.igraph.get_shortest_paths(root_vs[0], leaf_vs, mode="ALL")

        # Translate indices back into node ids
        ids = np.array(x.igraph.vs.get_attribute_values("node_id"))
        paths_tn = [ids[p] for p in paths]
    else:
        # Fall back to networkX
        iterator = nx.shortest_path(x.graph, target=root_nodes[0])
        paths_tn = [iterator[l][::-1] for l in leaf_nodes]

    # Generate DataFrame with all the info
    nodes = x.nodes.set_index("node_id")
    path_df = pd.DataFrame()
    path_df["path"] = paths_tn
    pdist = nodes.parent_dist.to_dict()
    path_df["distances"] = path_df.path.map(lambda x: np.array([pdist[n] for n in x]))
    path_df["cable"] = path_df.distances.map(lambda x: sum(x))

    # Sort DataFrame by cable length
    path_df.sort_values("cable", inplace=True, ascending=False)
    path_df.reset_index(inplace=True)

    # Prepare for plotting by finding starts points and defining angles
    positions = {x.root[0]: DEFAULTS["origin"]}
    angles = {x.root[0]: DEFAULTS["start_angle"]}

    seen = {x.root[0]}
    for k, path in enumerate(path_df.path.values):
        # Because the paths are always from tip to root, we have to find out
        # which of the nodes have already been plotted and at which branch point
        # we should add this path to the dendrogram
        path = np.asarray(path)
        exists = path[np.isin(path, list(seen))]
        n_branch_points = len(branch_nodes & set(exists))

        # Prune path to the bit that does not yet exist
        is_new = ~np.isin(path, list(seen))  # numpy.isin doesn't like sets
        path = path[is_new]
        start_point = positions[exists[-1]]
        last_angle = angles[exists[-1]]

        # Get distance of the remaining path
        distances = path_df.iloc[k].distances[is_new]
        distances[0] = 0
        distances = np.cumsum(distances)

        # Normalise distances
        if normalize_distance:
            longest_dist = path_df.iloc[0].cable
            distances /= longest_dist

        # Make sure the longest neurite goes horizontally
        # (or whatever START_ANGLE is)
        if k != 0:
            angle = DEFAULTS["angle_change"] - (
                DEFAULTS["angle_decrease"] * n_branch_points
            )
            angle = max(angle, DEFAULTS["angle_decrease"])
        else:
            angle = DEFAULTS["start_angle"]

        # Invert angle depending on odd or even branch points
        # (only to this if major branch -> SWITCH_DIST)
        if n_branch_points % 2 != 0 and distances[-1] >= DEFAULTS["switch_dist"]:
            angle *= -1

        # Angle to radians
        angle *= math.pi / 180

        # Add to last angle
        angle += last_angle

        # Calc x/y positions
        y_coords = np.array([math.sin(angle) * v for v in distances])
        x_coords = np.array([math.cos(angle) * v for v in distances])

        # Offset by starting point
        y_coords += start_point[1]
        x_coords += start_point[0]

        # Apply shade
        color = DEFAULTS["color"]
        if shade_by_length:
            a = 0.8 - 0.8 * distances[-1] / path_df.cable.max()
            color = mcl.to_rgba(color, alpha=a)

        # Change linewidths with path length
        lw = 1 * distances[-1] / path_df.cable.max() + 0.5

        # Plot
        ax.plot(
            x_coords, y_coords, color=color, zorder=path_df.shape[0] - k, linewidth=lw
        )

        # Keep track of positions for each treenode and angle of each path
        for i, coords in enumerate(zip(x_coords, y_coords)):
            positions[path[i]] = coords

            if path[i] not in angles:
                angles[path[i]] = angle

        seen = seen | set(path)

    # Plot connectors
    if connectors and x.has_connectors:
        # Get centers for each connector
        centers = np.vstack(x.connectors.node_id.map(positions))
        # Angle of the branch they belong to
        angles = (x.connectors.node_id.map(angles) + 90 * (math.pi / 180)).values

        # Create lines orthogonal to parent branch
        y_coords = np.sin(angles) * DEFAULTS["syn_marker_size"]
        y_coords = np.dstack(
            (
                y_coords + centers[:, 1],
                -y_coords + centers[:, 1],
                [None] * len(y_coords),
            )
        ).flatten()
        x_coords = np.cos(angles) * DEFAULTS["syn_marker_size"]
        x_coords = np.dstack(
            (
                x_coords + centers[:, 0],
                -x_coords + centers[:, 0],
                [None] * len(x_coords),
            )
        ).flatten()

        cn_cmap = prepare_connector_cmap(x)
        for ty in x.connectors.type.unique():
            is_type = x.connectors.type == ty
            is_type = np.dstack((is_type, is_type, is_type)).flatten()
            ax.plot(
                x_coords[is_type],
                y_coords[is_type],
                color=cn_cmap[ty]["color"],
                zorder=path_df.shape[0] + 1,
                linewidth=DEFAULTS["syn_linewidth"],
            )

    # Plot highlighted connectors
    if not isinstance(highlight_connectors, type(None)) and x.has_connectors:
        this = x.connectors[x.connectors.connector_id.isin(highlight_connectors)]
        # Get centers for each connector
        centers = np.vstack(this.node_id.map(positions))
        # Angle of the branch they belong to
        angles = (this.node_id.map(angles) + 90 * (math.pi / 180)).values

        # Create lines orthogonal to parent branch
        y_coords = np.sin(angles) * DEFAULTS["syn_marker_size"]
        y_coords = np.dstack(
            (
                y_coords + centers[:, 1],
                -y_coords + centers[:, 1],
                [None] * len(y_coords),
            )
        ).flatten()
        x_coords = np.cos(angles) * DEFAULTS["syn_marker_size"]
        x_coords = np.dstack(
            (
                x_coords + centers[:, 0],
                -x_coords + centers[:, 0],
                [None] * len(x_coords),
            )
        ).flatten()

        ax.plot(
            x_coords,
            y_coords,
            color=DEFAULTS["syn_highlight_color"],
            zorder=path_df.shape[0] + 2,
            linewidth=DEFAULTS["syn_linewidth"],
        )

    # Plot soma
    if x.has_soma:
        soma = utils.make_iterable(x.soma)[0]
        soma_pos = positions[soma]
        ax.scatter([soma_pos[0]], [soma_pos[1]], s=40, color=(0.1, 0.1, 0.1))

    # Make sure x/y axis are equal
    ax.set_aspect("equal")

    # Return axis
    return ax, positions


def _plot_force(
    x,
    connectors=False,
    highlight_connectors=None,
    prog="dot",
    ax=None,
    use_cache=True,
    **kwargs,
):
    """Plot neurons as dendrograms using graphviz layouts."""
    DEFAULTS = _DEFAULTS.copy()
    DEFAULTS.update(kwargs)
    # Save start time
    start = time.time()

    # Generate and populate networkX graph representation of the neuron
    G = x.graph.copy()
    # graphviz needs "len" not "weight"
    nx.set_edge_attributes(G, nx.get_edge_attributes(G, "weight"), name="len")

    # Calculate layout
    if x.core_md5 in _NODE_POSITIONS and use_cache:
        logger.info(
            "Using cached node positions. Set `use_cache=False` to "
            "calculate from scratch."
        )
        positions = _NODE_POSITIONS[x.core_md5]
    else:
        logger.info("Calculating node positions.")
        # Make sure this doesn't grow out of proportion
        while len(_NODE_POSITIONS) > 8:
            _ = _NODE_POSITIONS.popitem()

        positions = nx.nx_agraph.graphviz_layout(
            G, prog=prog, root=utils.make_iterable(x.soma)[0] if x.has_soma else None
        )
        _NODE_POSITIONS[x.core_md5] = positions

    # Plot tree with above layout
    logger.info("Plotting tree.")
    if not ax:
        fig, ax = plt.subplots(figsize=kwargs.get("figsize", (12, 6)))
        # Make background transparent (nicer for dark themes)
        fig.patch.set_alpha(0)
        ax.patch.set_alpha(0)

    nx.draw(
        G, positions, node_size=0, arrows=False, edge_color=DEFAULTS["color"], ax=ax
    )

    # Add soma
    if x.has_soma:
        for s in utils.make_iterable(x.soma):
            ax.scatter(
                [positions[s][0]],
                [positions[s][1]],
                s=40,
                color=DEFAULTS["color"],
                zorder=1,
            )

    if connectors and x.has_connectors:
        cn_cmap = prepare_connector_cmap(x)
        for ty in x.connectors.type.unique():
            this = x.connectors[x.connectors.type == ty]
            coords = np.vstack(this.node_id.map(positions))
            ax.scatter(
                coords[:, 0],
                coords[:, 1],
                color=cn_cmap[ty]["color"],
                zorder=2,
                s=DEFAULTS["syn_marker_size"] * 10,
            )

    if not isinstance(highlight_connectors, type(None)) and x.has_connectors:
        this = x.connectors[x.connectors.connector_id.isin(highlight_connectors)]
        if this.empty:
            logger.warning(
                "None of the connector IDs in `highlight_connectors`"
                "were found in neuron"
            )
        else:
            coords = np.vstack(this.node_id.map(positions))
            ax.scatter(
                coords[:, 0],
                coords[:, 1],
                color=DEFAULTS["syn_highlight_color"],
                zorder=3,
                s=DEFAULTS["syn_marker_size"] * 10,
            )

    logger.debug(f"Done in {time.time()-start}s")

    return ax, positions


def print_skeleton(x, print_func=print, add_props=None):
    """Print skeleton as hierarchical tree.

    Note: this really only makes sense for small neurons.

    Parameters
    ----------
    x :             TreeNeuron
                    Neuron to print.
    print_func :    callable, optional
                    Function to use for printing. Default is `print`.

    Returns
    -------
    None
    """

    def _print_tree(node, last=True, header=""):
        """Recursive print function."""
        elbow = "└──"
        pipe = "│  "
        tee = "├──"
        blank = "   "
        line = f"{header}{elbow if last else tee}{node}"
        if add_props:
            line += f" ({add_props[node]})"
        print_func(line)
        children = list(x.graph.predecessors(node))
        for i, c in enumerate(children):
            _print_tree(
                c,
                header=header + (blank if last else pipe),
                last=i == len(children) - 1,
            )

    if isinstance(add_props, str):
        add_props = x.nodes.set_index("node_id")[add_props].to_dict()
    elif isinstance(add_props, (list, np.ndarray, tuple)):
        if len(add_props) != len(x.nodes):
            raise ValueError("Length of add_props must be the same as number of")
        add_props = dict(zip(x.nodes.node_id, add_props))
    elif not isinstance(add_props, (dict, type(None))):
        raise ValueError(
            "add_props must be either None, a string, list, tuple or dict."
        )

    for root in x.root:
        _print_tree(root)



#    This script is part of navis (http://www.github.com/navis-org/navis).
#    Copyright (C) 2018 Philipp Schlegel
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

import functools

import numpy as np
import pandas as pd

from abc import ABC, abstractmethod
from inspect import signature

from .. import utils, config

logger = config.get_logger(__name__)


def trigger_init(func):
    """Trigger delayed initialization."""
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        self = args[0]
        # Check if has already been initialized
        if not self.initialized:
            self.__delayed_init__()
        return func(*args, **kwargs)
    return wrapper


class BaseTransform(ABC):
    """Abstract base class for transforms.

    If the transform is invertible, implement via __neg__ method.
    """

    def append(self, other: 'BaseTransform'):
        """Append another transform to this one.

        This is used to try to concatenate transforms of the same type into
        a single step to speed things up (e.g. for CMTK transforms). If that's
        not possible or not useful, must raise a `NotImplementedError`.
        """
        raise NotImplementedError(f'Unable to append {type(other)} to {type(self)}')

    def check_if_possible(self, on_error: str = 'raise'):
        """Test if running the transform is possible."""
        return

    @abstractmethod
    def copy(self) -> 'BaseTransform':
        """Return copy."""
        pass

    @abstractmethod
    def xform(self, points: np.ndarray) -> np.ndarray:
        """Return copy.

        Must accept a (N, 3) numpy array as first input and return the
        transformed (N, 3) points as sole output.s
        """
        pass


class AliasTransform(BaseTransform):
    """Helper transform that simply passes points through.

    Useful for defining aliases.
    """

    def __init__(self):
        """Initialize."""
        pass

    def __neg__(self) -> 'AliasTransform':
        """Invert transform."""
        return self.copy()

    def __eq__(self, other):
        """Check if the same."""
        if isinstance(other, AliasTransform):
            True
        return False

    def copy(self):
        """Return copy."""
        x = AliasTransform()
        x.__dict__.update(self.__dict__)
        return x

    def xform(self, points: np.ndarray) -> np.ndarray:
        """Pass through.

        Note that the returned points are NOT a copy but the originals.
        """
        return points


class FunctionTransform(BaseTransform):
    """Apply custom function as transform.

    Parameters
    ----------
    func :      callable
                Function that accepts and returns an (N, 3) array.

    """

    def __init__(self, func):
        """Initialize."""
        if not callable(func):
            raise TypeError('`func` must be callable')
        self.func = func

    def __eq__(self, other):
        """Check if the same."""
        if not isinstance(other, FunctionTransform):
            return False
        if self.func != other.func:
            return False
        return True

    def copy(self):
        """Return copy."""
        x = self.__class__(self.func)
        x.__dict__.update(self.__dict__)
        return x

    def xform(self, points: np.ndarray) -> np.ndarray:
        """Xform data.

        Parameters
        ----------
        points :        (N, 3) numpy array | pandas.DataFrame
                        Points to xform. DataFrame must have x/y/z columns.

        Returns
        -------
        pointsxf :      (N, 3) numpy array
                        Transformed points.

        """
        if isinstance(points, pd.DataFrame):
            # Make sure x/y/z columns are present
            if np.any([c not in points for c in ['x', 'y', 'z']]):
                raise ValueError('points DataFrame must have x/y/z columns.')
            points = points[['x', 'y', 'z']].values
        elif not (isinstance(points, np.ndarray) and points.ndim == 2 and points.shape[1] == 3):
            raise TypeError('`points` must be numpy array of shape (N, 3) or '
                            'pandas DataFrame with x/y/z columns')
        return self.func(points.copy())

class TransformSequence:
    """A sequence of transforms.

    Use this to apply multiple (different types of) transforms in sequence.

    Parameters
    ----------
    *transforms :   Transform/Sequences.
                    The transforms to bundle in this sequence.
    copy :          bool
                    Whether to make a copy of the transform on initialization.
                    This is highly recommended because otherwise we might alter
                    the original as we add more transforms (e.g. for CMTK
                    transforms).

    """

    def __init__(self, *transforms, copy=True):
        """Initialize."""
        self.transforms = []
        for tr in transforms:
            if not isinstance(tr, (BaseTransform, TransformSequence)):
                raise TypeError(f'Expected transform, got "{type(tr)}"')
            if copy:
                tr = tr.copy()
            self.append(tr)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return f'TransformSequence with {len(self)} transform(s)'

    def __len__(self) -> int:
        """Count number of transforms in this sequence."""
        return len(self.transforms)

    def __neg__(self) -> 'TransformSequence':
        """Invert transform sequence."""
        return TransformSequence(*[-t for t in self.transforms[::-1]])

    def append(self, transform: 'BaseTransform'):
        """Add transform to list."""
        if isinstance(transform, TransformSequence):
            # Unpack if other is sequence of transforms
            transform = transform.transforms

        for tr in utils.make_iterable(transform):
            if not isinstance(tr, BaseTransform):
                raise TypeError(f'Unable append "{type(tr)}"')

            if not hasattr(transform, 'xform') or not callable(transform.xform):
                raise TypeError('Transform does not appear to have a `xform` method')

            # Try to merge with the last transform in the sequence
            if len(self):
                try:
                    self.transforms[-1].append(tr)
                except NotImplementedError:
                    self.transforms.append(tr)
                except BaseException:
                    raise
            else:
                self.transforms.append(tr)

    def xform(self, points: np.ndarray,
              affine_fallback: bool = True,
              **kwargs) -> np.ndarray:
        """Perform transforms in sequence."""
        # First check if any of the transforms raise any issues ahead of time
        # This can e.g. be missing binaries like CMTK's streamxform
        for tr in self.transforms:
            tr.check_if_possible(on_error='raise')

        # Now transform points in sequence
        # Make a copy of the points to avoid changing the originals
        # Note dtype float64 in case our precision in case precisio must go up
        # -> e.g. when converting from nm to micron space
        xf = np.asarray(points).astype(np.float64)
        for tr in self.transforms:
            # Check this transforms signature for accepted Parameters
            params = signature(tr.xform).parameters

            # We must not pass NaN value from one transform to the next
            is_nan = np.any(np.isnan(xf), axis=1)

            # Skip if all points are NaN
            if all(is_nan):
                continue

            if 'affine_fallback' in params:
                xf[~is_nan] = tr.xform(xf[~is_nan],
                                       affine_fallback=affine_fallback,
                                       **kwargs)
            else:
                xf[~is_nan] = tr.xform(xf[~is_nan], **kwargs)

        return xf


class TransOptimizer:
    """Optimizes a Transform or TransformSequence.

    The purpose of this class is to change a bunch of settings (depending on the
    type of transforms) before running the transformation and do a clean up
    after it has finished.

    Currently, the optimizations are very hands-on but in the future, this might
    be delegated to the individuals `Transform` classes - e.g. by implementing
    an `.optimize()` method which is then called by `TransOptimizer`.

    Currently, it really only manages caching for H5 transforms.

    Parameters
    ----------
    tr :        Transform | TransformSequence
                The transform or sequence thereof to be optimized.
    mode :      None | "medium" | "aggressive"
                Mode for optimization:
                  - `None`: no optimization
                  - "medium": some optimization but keep upfront cost low
                  - "aggressive": high upfront cost but should be faster in the long run

    Examples
    --------
    >>> from navis.transforms import h5reg
    >>> from navis.transforms.base import TransOptimizer
    >>> tr = h5reg.H5transform('path/to/reg.h5', direction='inverse') # doctest: +SKIP
    >>> with TransOptimizer(tr, mode='aggressive'):                   # doctest: +SKIP
    >>>     xf = tr.xform(pts)                                        # doctest: +SKIP

    """

    def __init__(self, tr, bbox, caching: bool):
        """Initialize Optimizer."""
        assert isinstance(caching, bool)

        self.caching = caching
        self.bbox = np.asarray(bbox)

        assert self.bbox.ndim == 2 and self.bbox.shape == (3, 2)

        if isinstance(tr, BaseTransform):
            self.transforms = [tr]
        elif isinstance(tr, TransformSequence):
            self.transforms = tr.transforms
        else:
            raise TypeError(f'Expected Transform/Sequence, got "{type(tr)}"')

    def __enter__(self):
        """Apply optimizations."""
        if not self.caching:
            return

        # Check if there are any transforms we can optimize
        if not any(['H5transform' in str(type(tr)) for tr in self.transforms]):
            return

        if not config.pbar_hide:
            logger.info('Pre-caching deformation field(s) for transforms...')

        bbox_xf = self.bbox
        for tr in self.transforms:
            # We are monkey patching here to avoid circular imports
            # not pretty but works
            if 'H5transform' in str(type(tr)):
                # Precache values in the bounding box
                tr.precache(bbox_xf, padding=True)
            # To pre-cache sequential transforms we need to xform the bounding
            # box as we move along
            bbox_xf = tr.xform(bbox_xf.T).T

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Revert optimizations."""
        if not self.caching:
            return

        for tr in self.transforms:
            # We are monkey patching here to avoid circular imports
            # not pretty but works
            if 'H5transform' in str(type(tr)):
                # Clears the cache
                tr.use_cache = False
